"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.SemigroupSum = exports.SemigroupMultiply = exports.SemigroupMin = exports.SemigroupMax = exports.MonoidSum = exports.MonoidMultiply = void 0;
var _BigInt = require("effect/BigInt");
var monoid = _interopRequireWildcard(require("../Monoid.js"));
var semigroup = _interopRequireWildcard(require("../Semigroup.js"));
function _interopRequireWildcard(e, t) { if ("function" == typeof WeakMap) var r = new WeakMap(), n = new WeakMap(); return (_interopRequireWildcard = function (e, t) { if (!t && e && e.__esModule) return e; var o, i, f = { __proto__: null, default: e }; if (null === e || "object" != typeof e && "function" != typeof e) return f; if (o = t ? n : r) { if (o.has(e)) return o.get(e); o.set(e, f); } for (const t in e) "default" !== t && {}.hasOwnProperty.call(e, t) && ((i = (o = Object.defineProperty) && Object.getOwnPropertyDescriptor(e, t)) && (i.get || i.set) ? o(f, t, i) : f[t] = e[t]); return f; })(e, t); }
/**
 * @since 0.24.0
 */

/**
 * `bigint` semigroup under addition.
 *
 * **Example**
 *
 * ```ts
 * import { SemigroupSum } from "@effect/typeclass/data/BigInt"
 *
 * console.log(SemigroupSum.combine(2n, 3n))
 * // 5n
 * ```
 *
 * @category instances
 * @since 0.24.0
 */
const SemigroupSum = exports.SemigroupSum = /*#__PURE__*/semigroup.make((self, that) => self + that);
/**
 * `bigint` semigroup under multiplication.
 *
 * @category instances
 * @since 0.24.0
 */
const SemigroupMultiply = exports.SemigroupMultiply = /*#__PURE__*/semigroup.make((self, that) => self * that, (self, collection) => {
  if (self === 0n) {
    return 0n;
  }
  let out = self;
  for (const n of collection) {
    if (n === 0n) {
      return 0n;
    }
    out = out * n;
  }
  return out;
});
/**
 * A `Semigroup` that uses the minimum between two values.
 *
 * **Example**
 *
 * ```ts
 * import { SemigroupMin } from "@effect/typeclass/data/BigInt"
 *
 * console.log(SemigroupMin.combine(2n, 3n))
 * // 2n
 * ```
 *
 * @category instances
 * @since 0.24.0
 */
const SemigroupMin = exports.SemigroupMin = /*#__PURE__*/semigroup.min(_BigInt.Order);
/**
 * A `Semigroup` that uses the maximum between two values.
 *
 * **Example**
 *
 * ```ts
 * import { SemigroupMax } from "@effect/typeclass/data/BigInt"
 *
 * console.log(SemigroupMax.combine(2n, 3n))
 * // 3n
 * ```
 *
 * @category instances
 * @since 0.24.0
 */
const SemigroupMax = exports.SemigroupMax = /*#__PURE__*/semigroup.max(_BigInt.Order);
/**
 * `bigint` monoid under addition.
 *
 * The `empty` value is `0n`.
 *
 * **Example**
 *
 * ```ts
 * import { MonoidSum } from "@effect/typeclass/data/BigInt"
 *
 * console.log(MonoidSum.combine(2n, 3n))
 * // 5n
 * console.log(MonoidSum.combine(2n, MonoidSum.empty))
 * // 2n
 * ```
 *
 * @category instances
 * @since 0.24.0
 */
const MonoidSum = exports.MonoidSum = /*#__PURE__*/monoid.fromSemigroup(SemigroupSum, 0n);
/**
 * `bigint` monoid under multiplication.
 *
 * The `empty` value is `1n`.
 *
 * **Example**
 *
 * ```ts
 * import { MonoidMultiply } from "@effect/typeclass/data/BigInt"
 *
 * console.log(MonoidMultiply.combine(2n, 3n))
 * // 6n
 * console.log(MonoidMultiply.combine(2n, MonoidMultiply.empty))
 * // 2n
 * ```
 *
 * @category instances
 * @since 0.24.0
 */
const MonoidMultiply = exports.MonoidMultiply = /*#__PURE__*/monoid.fromSemigroup(SemigroupMultiply, 1n);
//# sourceMappingURL=BigInt.js.map